// Copyright 2019 dart-chacha20 authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import 'dart:typed_data';

import 'package:benchmark_harness/benchmark_harness.dart';
import 'package:chacha20/chacha20.dart';

Chacha20 chacha20 = Chacha20();

const million = 1000000;

void main() {
  print("Benchmarks:");
  print("--------------");
  print("NORMAL VERSION");
  print("--------------");
  print("--");
  Chacha20BlockBenchmark(million).report();
  print("--");
  Chacha20StreamBenchmark(million).report();
  print("--");
  Chacha20NumerousSmallMessagesBenchmark(million, 100).report();
  print("--");

  print("-------------------");
  print("NON-INLINED VERSION");
  print("-------------------");
  chacha20 = NonInlinedChacha20();
  Chacha20BlockBenchmark(million).report();
  print("--");
  Chacha20StreamBenchmark(million).report();
  print("--");
  Chacha20NumerousSmallMessagesBenchmark(million, 100).report();
  print("--");
}

class Chacha20BlockBenchmark extends BenchmarkBase {
  final int totalLength;
  Uint32List state;

  Chacha20BlockBenchmark(this.totalLength)
      : super("${totalLength ~/ million} MB (blocks only, no xor)");

  @override
  void setup() {
    chacha20.initialize(key: Uint8List(32), nonce: null);
    state = Uint32List(16);
  }

  @override
  void run() {
    chacha20.encryptState(state);
  }

  @override
  void exercise() {
    for (var i = 0; i < totalLength ~/ 64; i++) {
      run();
    }
  }
}

class Chacha20StreamBenchmark extends BenchmarkBase {
  final int totalLength;
  Uint8List cleartext;
  Uint8List result;

  Chacha20StreamBenchmark(this.totalLength)
      : super("${totalLength ~/ million} MB stream");

  @override
  void setup() {
    // 100 MB cleartext
    cleartext = Uint8List(totalLength);
    for (var i = 0; i < cleartext.lengthInBytes; i++) {
      cleartext[i] = 0xFF & i;
    }
    result = Uint8List(cleartext.length);
  }

  @override
  void run() {
    chacha20.initialize(key: Uint8List(32), nonce: null);
    chacha20.fillWithConverted(result, 0, cleartext, 0);
  }

  @override
  void exercise() {
    run();
  }
}

class Chacha20NumerousSmallMessagesBenchmark extends BenchmarkBase {
  final int totalLength;
  final int messageLength;
  Uint8List cleartext;
  Uint8List result;

  Chacha20NumerousSmallMessagesBenchmark(this.totalLength, this.messageLength)
      : super(
            "${totalLength ~/ million} MB in ${messageLength} byte long messages");

  @override
  void setup() {
    cleartext = Uint8List(messageLength);
    for (var i = 0; i < cleartext.lengthInBytes; i++) {
      cleartext[i] = 0xFF & i;
    }
    result = Uint8List(cleartext.lengthInBytes);
  }

  @override
  void run() {
    chacha20.initialize(key: Uint8List(32), nonce: null);
    chacha20.fillWithConverted(result, 0, cleartext, 0);
  }

  @override
  void exercise() {
    for (var i = 0; i < totalLength ~/ messageLength; i++) {
      run();
    }
  }
}

/// Non-inlined implementation.
/// Left here so developers can benchmark it.
class NonInlinedChacha20 extends Chacha20 {
  @override
  void encryptState(Uint32List state) {
    final initialState = this.initialState;

    // -------------------------------------------------------------------------
    // Step 1: Initialize
    // -------------------------------------------------------------------------
    for (var i = 0; i < state.length; i++) {
      state[i] = initialState[i];
    }

    // -------------------------------------------------------------------------
    // Step 2: Do 20 column/diagonal rounds
    // -------------------------------------------------------------------------
    for (var i = 0; i < 10; i++) {
      // A column round
      quarterRound(state, 0, 4, 8, 12);
      quarterRound(state, 1, 5, 9, 13);
      quarterRound(state, 2, 6, 10, 14);
      quarterRound(state, 3, 7, 11, 15);

      // A diagonal round
      quarterRound(state, 0, 5, 10, 15);
      quarterRound(state, 1, 6, 11, 12);
      quarterRound(state, 2, 7, 8, 13);
      quarterRound(state, 3, 4, 9, 14);
    }

    // -------------------------------------------------------------------------
    // Step 3: Addition
    // -------------------------------------------------------------------------
    for (var i = 0; i < state.length; i++) {
      state[i] = _uint32mask & (state[i] + initialState[i]);
    }
  }

  void quarterRound(Uint32List state, int a, int b, int c, int d) {
    int v0 = state[a];
    int v4 = state[b];
    int v8 = state[c];
    int v12 = state[d];
    v0 = _uint32mask & (v0 + v4);
    v12 = _rotateLeft(v12 ^ v0, 16);
    v8 = _uint32mask & (v8 + v12);
    v4 = _rotateLeft(v4 ^ v8, 12);
    v0 = _uint32mask & (v0 + v4);
    v12 = _rotateLeft(v12 ^ v0, 8);
    v8 = _uint32mask & (v8 + v12);
    v4 = _rotateLeft(v4 ^ v8, 7);
    state[a] = v0;
    state[b] = v4;
    state[c] = v8;
    state[d] = v12;
  }

  /// For ensuring that result of calculation will be 32-bit integer.
  static const int _uint32mask = 0xFFFFFFFF;

  /// Rotates 32-bit integer to the left.
  static int _rotateLeft(int data, int shift) {
    return _uint32mask & ((data << shift) | (data >> (32 - shift)));
  }
}
